# `delay_samples` param is not reliably optimizable. See `delay_line.py` for more details.

import numpy as np
import jax.numpy as jnp
from jax import jit

from jaxdsp.param import Param

MAX_DELAY_SIZE_SAMPLES = 44100

NAME = "Feedforward Delay"
PARAMS = [
    Param("wet", 1.0),
    Param("delay_samples", 6.5, 0.0, float(MAX_DELAY_SIZE_SAMPLES)),
]
PRESETS = {}


def init_state():
    return {}


def tick(carry, x):
    raise "single-sample tick method not implemented for feedforward_delay"


@jit
def tick_buffer(carry, X):
    params, _ = carry
    delay_samples = params["delay_samples"]
    remainder = delay_samples - jnp.floor(delay_samples)
    X_linear_interp = (1 - remainder) * X + remainder * jnp.concatenate(
        [jnp.array([0]), X[: X.size - 1]]
    )
    Y = jnp.where(
        jnp.arange(X.size) >= delay_samples,
        jnp.roll(X_linear_interp, delay_samples.astype("int32")),
        0,
    )
    return carry, X * (1 - params["wet"]) + Y * params["wet"]



from jaxdsp.processors import (
    allpass_filter,
    clip,
    delay_line,
    feedforward_delay,
    fir_filter,
    freeverb,
    iir_filter,
    lowpass_feedback_comb_filter,
    sine_wave,
)

all_processors = [
    allpass_filter,
    clip,
    delay_line,
    feedforward_delay,
    fir_filter,
    freeverb,
    iir_filter,
    lowpass_feedback_comb_filter,
    sine_wave,
]
processor_by_name = {processor.NAME: processor for processor in all_processors}


def default_param_values(processor):
    return (
        {param.name: param.default_value for param in processor.PARAMS}
        if processor
        else None
    )


def id_for_processor(processor):
    return all_processors.index(processor)


def serialize_processor(processor, params=None):
    if not processor:
        return None

    return {
        "name": processor.NAME,
        "param_definitions": [param.serialize() for param in processor.PARAMS],
        "presets": processor.PRESETS,
        "params": params or default_param_values(processor),
    }


def params_to_unit_scale(params, processor_name):
    processor = processor_by_name[processor_name]
    return (
        {
            param.name: param.to_unit_scale(params[param.name])
            for param in processor.PARAMS
        }
        if processor
        else {}
    )


def params_from_unit_scale(params, processor_name):
    processor = processor_by_name[processor_name]
    return (
        {
            param.name: param.from_unit_scale(params[param.name])
            for param in processor.PARAMS
        }
        if processor
        else {}
    )

# Returns (params, state)
def processor_config_to_carry(processor_config):
    if not processor_config:
        return None, None
    params = [
        processor["params"]
        if "params" in processor
        else default_param_values(processor_by_name[processor["name"]])
        for processor in processor_config
    ]
    state = [
        processor_by_name[processor["name"]].init_state()
        for processor in processor_config
    ]
    return params, state



import jax.numpy as jnp

from jaxdsp.processors import processor_by_name

# This does not actually support full graph connectivity.
# For simplicity, only series & parallel processing is supported.
# A wide variety of common DSP techniques can be implemented by nesting series and parallel processor groups.
# Keeping this restriction (at least for now) means the graph structure can be represented and serialized
# simply as a list, where each element can be a processor or a list.
# This also simplifies the UI since the "graph" connectivity is implicit in the positional order.
# Thus, we don't need to support arbitrary connectivity, and can instead use a simple drag-and-drop UX paradigm.

def tick_buffer_series(carry, X, processor_names):
    params, state = carry

    Y = X
    for i, processor_name in enumerate(processor_names):
        processor_state = state[i]
        processor_params = params[i]
        processor = processor_by_name[processor_name]
        processor_carry, Y = processor.tick_buffer(
            (processor_params, processor_state), Y
        )
        carry[1][i] = processor_carry[1]

    return carry, Y


def tick_buffer_parallel(carry, X, processor_names):
    params, state = carry

    Y = jnp.zeros(X.shape)
    for i, processor_name in enumerate(processor_names):
        processor_state = state[i]
        processor_params = params[i]
        processor = processor_by_name[processor_name]
        processor_carry, Y_i = processor.tick_buffer(
            (processor_params, processor_state), X
        )
        Y += Y_i
        carry[1][i] = processor_carry[1]

    return carry, Y



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="jaxdsp",
    version="0.2.0",
    author="Karl Hiner",
    author_email="karl.hiner@gmail.com",
    description="Fast, differentiable audio processors on the CPU or GPU",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/khiner/jaxdsp",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
